"""
Interval definitions for the subsequent model fitting.
"""

import numpy as np

from abc import ABC, abstractmethod

__all__ = [
    "WidthOfIntervalSlicer",
    "NumberOfIntervalsSlicer",
    "PointsPerIntervalSlicer",
]


class IntervalSlicer(ABC):
    """
    Abstract base class for IntervalSlicer

    Sorts the conditional variable (e.g Tp|Hs) into intervals of the
    independent variable (Hs).

    """

    def __init__(self, **kwargs):
        # check if there are unknown kwargs
        kwarg_keys = kwargs.keys()
        unknown_kwarg_keys = set(kwarg_keys).difference(
            {"min_n_intervals", "min_n_points"}
        )
        if len(unknown_kwarg_keys) != 0:
            raise TypeError(
                "__init__() got an unexpected keyword argument "
                f"'{unknown_kwarg_keys.pop()}'"
            )

        self.min_n_points = kwargs.get("min_n_points", 50)
        self.min_n_intervals = kwargs.get("min_n_intervals", 3)
        self.reference = None

    def slice_(self, data):
        """
        Slices the data into intervals of equal width.

        Parameters
        ----------
        data : one-dimensional ndarray.
            Contains the data of the independent variable.

        Returns
        -------
        interval_slices: list of ndarray
            Boolean arrays with same length as data. One for each interval.
            True where a value in data falls in the corresponding interval.

        interval_references: ndarray
            Reference points of intervals. Length equal to number of intervals.

        interval_boundaries: list of tuple
            List of (upper, lower) limit tuples. One tuple for each interval.

        """

        interval_slices, interval_references, interval_boundaries = self._slice(data)

        if len(interval_slices) < self.min_n_intervals:
            raise RuntimeError(
                "Slicing resulting in too few intervals. "
                f"Need at least {self.min_n_intervals}, "
                f"but got only {len(interval_slices)} intervals."
            )

        if callable(self.reference):
            interval_references = [
                self.reference(data[slice_]) for slice_ in interval_slices
            ]

        return interval_slices, interval_references, interval_boundaries

    @abstractmethod
    def _slice(self, data):
        pass

    def _drop_too_small_intervals(
        self, interval_slices, interval_references, interval_boundaries
    ):
        ok_slices = []
        ok_references = []
        ok_boundaries = []
        for slice_, int_cent, int_bounds in zip(
            interval_slices, interval_references, interval_boundaries
        ):
            # slice_ is a boolean array, so sum returns number of points in interval
            if np.sum(slice_) >= self.min_n_points:
                ok_slices.append(slice_)
                ok_references.append(int_cent)
                ok_boundaries.append(int_bounds)
        return ok_slices, ok_references, ok_boundaries


class WidthOfIntervalSlicer(IntervalSlicer):
    """
    IntervalSlicer that uses width of intervals to define intervals.

    Parameters
    ----------
    width : float
        The width of each interval.
    reference : str or callable, optional
        Determines the reference value for each interval.
        If a string the following keywords are available:
        'center': use the center / midpoint of the interval as reference,
        'left': use the left / lower bound of the interval and
        'right': use the right / upper bound of the interval as reference.
        If a callable, a function is expected, that maps from an array with
        the values of an interval to the reference of that interval
        (e.g. np.median). Defaults to 'center'.
    right_open : boolean, optional
        Determines how the boundaries of the intervals are defined. Either
        the left or the right boundary is inclusive. Defaults to True,
        meaning the left boundary is inclusive and the right exclusive,
        i.e. :math:`[a, b)`.
    value_range : tuple, optional
        Determines the value range used for creating the intervals.
        If None, 0 and np.max(data) are used.
        If a 2-tuple it contains the lower and upper limit of the range.
        If either entry of the tuple is None the default for that entry is
        used. Defaults to None.
    min_n_points : int, optional
        Minimal number of points per interval. Intervals with fewer points
        are discarded. Defaults to 50.
    min_n_intervals : int, optional
        Minimal number of intervals. Raises a RuntimeError if slicing
        resulted in fewer intervals. Defaults to 3.

    Raises
    ------
    RuntimeError
        if slicing resulted in fewer than min_n_intervals intervals.
    """

    def __init__(
        self, width, reference="center", right_open=True, value_range=None, **kwargs
    ):
        super().__init__(**kwargs)
        self.width = width
        self.reference = reference
        self.right_open = right_open
        self.value_range = value_range

    def _slice(self, data):
        if self.value_range is None:
            data_min = 0
            data_max = np.max(data)
        else:
            if self.value_range[0] is not None:
                data_min = self.value_range[0]
            else:
                data_min = 0
            if self.value_range[1] is not None:
                data_max = self.value_range[1]
            else:
                data_max = np.max(data)

        width = self.width
        interval_references = np.arange(data_min, data_max + width, width) + 0.5 * width

        if self.right_open:
            interval_slices = [
                ((int_cent - 0.5 * width <= data) & (data < int_cent + 0.5 * width))
                for int_cent in interval_references
            ]
        else:
            interval_slices = [
                ((int_cent - 0.5 * width < data) & (data <= int_cent + 0.5 * width))
                for int_cent in interval_references
            ]

        interval_boundaries = [
            (c - width / 2, c + width / 2) for c in interval_references
        ]

        if isinstance(self.reference, str):
            if self.reference.lower() == "center":
                pass  # interval_references are already center of intervals
            elif self.reference.lower() == "right":
                interval_references += 0.5 * width
            elif self.reference.lower() == "left":
                interval_references -= 0.5 * width
            else:
                raise ValueError(
                    "Unknown value for 'reference'. "
                    "Supported values are 'center', 'left', "
                    f"and 'right', but got '{self.reference}'."
                )
        elif callable(self.reference):
            pass  #  handled in super class
        else:
            raise TypeError(
                "Wrong type for reference. Expected str or callable, "
                f"but got {type(self.reference)}."
            )

        (
            interval_slices,
            interval_references,
            interval_boundaries,
        ) = self._drop_too_small_intervals(
            interval_slices, interval_references, interval_boundaries
        )

        return interval_slices, interval_references, interval_boundaries


class NumberOfIntervalsSlicer(IntervalSlicer):
    """
    IntervalSlicer that uses a number of intervals to define intervals of
    equal width.

    Parameters
    ----------
    n_intervals : int
        Number of intervals the dataset is split into.
    reference : str or callable, optional
        Determines the reference value for each interval.
        If a string the following keywords are available:
        'center': use the center / midpoint of the interval as reference,
        'left': use the left / lower bound of the interval and
        'right': use the right / upper bound of the interval as reference.
        If a callable, a function is expected, that maps from an array with
        the values of an interval to the reference of that interval
        (e.g. np.median). Defaults to 'center'.
    include_max : boolean, optional
        Determines if the upper boundary of the last interval is inclusive.
        True if inclusive. Defaults to True.
    value_range : tuple or None, optional
        Determines the value range used for creating n_intervals equally
        sized intervals. If a tuple it contains the upper and lower limit
        of the range. If None the min and max of the data are used.
        Defaults to None.
    min_n_points : int, optional
        Minimal number of points per interval. Intervals with fewer points
        are discarded. Defaults to 50.
    min_n_intervals : int, optional
        Minimal number of intervals. Raises a RuntimeError if slicing
        resulted in fewer intervals. Defaults to 3.

    Raises
    ------
    RuntimeError
        if slicing resulted in fewer than min_n_intervals intervals.
    """

    def __init__(
        self,
        n_intervals,
        reference="center",
        include_max=True,
        value_range=None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        if n_intervals < self.min_n_intervals:
            self.min_n_intervals = n_intervals
        self.n_intervals = n_intervals
        self.reference = reference
        self.include_max = include_max
        self.value_range = value_range

    def _slice(self, data):
        if self.value_range is not None:
            value_range = self.value_range
        else:
            value_range = (min(data), max(data))

        interval_starts, interval_width = np.linspace(
            value_range[0],
            value_range[1],
            num=self.n_intervals,
            endpoint=False,
            retstep=True,
        )
        interval_references = interval_starts + 0.5 * interval_width

        interval_boundaries = [
            (c - interval_width / 2, c + interval_width / 2)
            for c in interval_references
        ]

        if isinstance(self.reference, str):
            if self.reference.lower() == "center":
                pass  # default
            elif self.reference.lower() == "right":
                interval_references = interval_starts + interval_width
            elif self.reference.lower() == "left":
                interval_references = interval_starts
            else:
                raise ValueError(
                    "Unknown value for 'reference'. "
                    "Supported values are 'center', 'left', "
                    f"and 'right', but got '{self.reference}'."
                )
        elif callable(self.reference):
            pass  #  handled in super class
        else:
            raise TypeError(
                "Wrong type for reference. Expected str or callable, "
                f"but got {type(self.reference)}."
            )

        interval_slices = [
            ((data >= int_start) & (data < int_start + interval_width))
            for int_start in interval_starts[:-1]
        ]

        # include max in last interval ?
        int_start = interval_starts[-1]
        if self.include_max:
            interval_slices.append(
                ((data >= int_start) & (data <= int_start + interval_width))
            )
        else:
            interval_slices.append(
                ((data >= int_start) & (data < int_start + interval_width))
            )

        (
            interval_slices,
            interval_references,
            interval_boundaries,
        ) = self._drop_too_small_intervals(
            interval_slices, interval_references, interval_boundaries
        )

        return interval_slices, interval_references, interval_boundaries


class PointsPerIntervalSlicer(IntervalSlicer):
    """
    Uses a number of points per interval to define intervals.

    Sorts the data and splits it into intervals with the same number of
    points. In general this results in intervals with varying width.

    Parameters
    ----------
    n_points : int
        The number of points per interval.
    reference : callable, optional
        Determines the reference value for each interval.
        A function is expected, that maps from an array with
        the values of an interval to the reference of that interval.
        Defaults to np.median.
    last_full : boolean, optional
        If it is not possible to split the data in chunks with the same
        number of points, one interval will have fewer points. This
        determines if the last or the first interval should have n_points
        points. If True the last interval contains n_points points and the
        first interval contains the remaining points. Defaults to True.
    min_n_points : int, optional
        Minimal number of points per interval. Intervals with fewer points
        are discarded. Defaults to 50.
    min_n_intervals : int, optional
        Minimal number of intervals. Raises a RuntimeError if slicing
        resulted in fewer intervals. Defaults to 3.

    Raises
    ------
    RuntimeError
        if slicing resulted in fewer than min_n_intervals intervals.
    """

    def __init__(self, n_points, reference=np.median, last_full=True, **kwargs):
        super().__init__(**kwargs)
        if n_points < self.min_n_points:
            self.min_n_points = n_points

        self.n_points = n_points
        self.reference = reference
        self.last_full = last_full

    def _slice(self, data):
        sorted_idc = np.argsort(data)
        n_full_chunks = len(data) // self.n_points
        remainder = len(data) % self.n_points
        if remainder != 0:
            if self.last_full:
                interval_idc = np.split(sorted_idc[remainder:], n_full_chunks)
                interval_idc.insert(0, sorted_idc[:remainder])

            else:
                interval_idc = np.split(
                    sorted_idc[: len(data) - remainder], n_full_chunks
                )
                interval_idc.append(sorted_idc[len(data) - remainder :])
        else:
            interval_idc = np.split(sorted_idc, n_full_chunks)

        interval_slices = [
            np.isin(sorted_idc, idc, assume_unique=True) for idc in interval_idc
        ]
        interval_references = [None] * len(
            interval_slices
        )  # gets overwritten in super().slice_ anyway

        # Pass interval_references twice instead of boundaries. We calculate
        # boundaries later.
        interval_slices, interval_references, _ = self._drop_too_small_intervals(
            interval_slices, interval_references, interval_references
        )

        # calculate the interval boundaries
        # the boundary between two intervals shall be the mean of
        # the max of the lower interval and the min of the higher interval
        # for the first interval the lower limit is the min of the data in that interval
        # for the last interval the upper limit is the max of the data in that interval
        interval_boundaries = []
        lower_boundary = np.min(data[interval_slices[0]])
        interval = data[interval_slices[0]]
        for i in range(len(interval_slices) - 1):
            # calculate boundaries for ith interval
            next_interval = data[interval_slices[i + 1]]
            upper_boundary = (np.max(interval) + np.min(next_interval)) / 2
            interval_boundaries.append((lower_boundary, upper_boundary))
            # prepare variables for next interval
            lower_boundary = upper_boundary
            interval = next_interval

        # append boundaries for last interval
        upper_boundary = np.max(interval)
        interval_boundaries.append((lower_boundary, upper_boundary))

        return interval_slices, interval_references, interval_boundaries



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Sample almost equally distributed points from an n-dimensional hypersphere.
"""

import itertools

import numpy as np

__all__ = ["NSphere"]


class NSphere:
    """
    This class calculates almost equally spaced points on a n-sphere.

    It considers the Thomson problem [#]_ to distribute the points on the sphere.

    References
    ----------
    .. [#] https://en.wikipedia.org/wiki/Thomson_problem

    Examples
    --------

    >>> sphere = NSphere(3, 1000)
    >>> samples = sphere.unit_sphere_points
    >>> #%% plot sphere
    >>> from mpl_toolkits.mplot3d import Axes3D
    >>> import matplotlib.pyplot as plt
    >>> plt.close('all')
    >>> #fig = plt.figure(figsize=plt.figaspect(1))
    >>> #ax = fig.add_subplot(111, projection='3d')
    >>> u = np.linspace(0, 2 * np.pi, 50)
    >>> v = np.linspace(0, np.pi, 30)
    >>> x = np.outer(np.cos(u), np.sin(v))
    >>> y = np.outer(np.sin(u), np.sin(v))
    >>> z = np.outer(np.ones(np.size(u)), np.cos(v))
    >>> #wireframe_plot = ax.plot_wireframe(x, y, z, rcount=10, ccount=10,)
    >>> #sphere_plot = ax.scatter(samples[:, 0], samples[:, 1], samples[:, 2],\
                                 c='green', depthshade=False, label="verteilt")
    >>> #plot_legend = ax.legend()

    """

    def __init__(self, dim, n_samples):
        """
        Parameters
        ----------
        dim : int
            The number of dimensions. (i.e. the n in n-sphere plus 1)
        n_samples : int
            The number of points to distribute on the n-sphere.

        """
        self.dim = dim
        self.n_samples = n_samples

        # combinations for selecting all different indice
        self.combs = np.array(
            [comb for comb in itertools.combinations(range(self.n_samples), 2)]
        )

        self.unit_sphere_points = self._random_unit_sphere_points()
        self.init_e_pot = self._pot_energy()
        self._relax_points()
        self.rela_e_pot = self._pot_energy()
        self.improvement = (self.init_e_pot - self.rela_e_pot) / self.init_e_pot * 100

    def _relax_points(
        self,
    ):
        """
        Iteratively move points to reduce potential energy.

        Iteratively moves points in the direction of the tangential part of the
        coloumb forces and calculates the potential energy after each iteration.
        The best state, i.e. the state with least potential energy, is saved.
        """

        best_state = np.copy(self.unit_sphere_points)
        best_pot = self._pot_energy()

        # Define the number osf iterations based on the size of the sample size.
        # At least 10 iterations should be performed. 10 is an empirical value.
        max_iters = max([10, int(10000 / self.n_samples)])

        for iteration in range(1, max_iters):
            # Tau controls the step size of the optimization. With each iteration
            # a smaller step is chosen. 3 is an empirical value.
            tau = 3 / iteration

            tang_forces = self._tangential_forces(self._get_forces())

            # Norm to max force
            max_force = np.max(
                np.linalg.norm(
                    tang_forces,
                    axis=1,
                )
            )
            tang_forces /= max_force

            self.unit_sphere_points += tang_forces * tau
            self.unit_sphere_points /= np.linalg.norm(
                self.unit_sphere_points, axis=1, keepdims=True
            )

            curr_pot = self._pot_energy()
            if curr_pot < best_pot:
                best_state = np.copy(self.unit_sphere_points)
                best_pot = curr_pot

        self.unit_sphere_points = best_state

    def _random_unit_sphere_points(self):
        """
        Generates equally distributed points on the sphere's surface.

        Note
        ----
        Used algorithm:

        - Use a N(0,1) standard normal distribution to generate cartesian coordinate vectors.
        - Normalize the vectors to length 1.
        - The points then are uniformly distributed on the sphere's surface.

        """
        # create pseudorandom number generator with seed for reproducability
        prng = np.random.RandomState(seed=43)
        #  draw normally distributed samples
        rand_points = prng.normal(size=(self.n_samples, self.dim))
        # calculate lengths of vectors
        radii = np.linalg.norm(rand_points, axis=1, keepdims=True)
        # normalize
        return rand_points / radii

    def _pot_energy(
        self,
    ):
        """
        Calculates the potential energy of the current state.

        Assume the points on the sphere are electrons with charge equal to 1 and
        assume Coulomb's constant equal to 1. Then the electrostatic potential energy of
        :math:`N` electrons can be written as:

        .. math::

            U = \\sum_{i < j } \\frac{1}{\\lvert {r_{ij}} \\rvert}

        With :math:`0 \\leq i < j \\leq N` and :math:`\\lvert {r_{ij}} \\rvert` the distance vector
        between electron :math:`i` and :math:`j`.

        """

        dist_vectors = (
            self.unit_sphere_points[self.combs[:, 0]]
            - self.unit_sphere_points[self.combs[:, 1]]
        )

        distances = np.linalg.norm(dist_vectors, axis=1, keepdims=True)

        return np.sum(distances**-1)

    def _get_forces(
        self,
    ):
        """
        Calculates the Coloumb forces.

        Assume the :math:`N` points are electrons with charge 1 and Coloumb's constant is 1.
        The Coulomb force on point :math:`i` can then be expressed as:

        .. math::

            F_i = \\sum_{j \\neq i} \\frac{1}{{\\lvert {r_{ij}} \\rvert}^2},
            0 \\leq j \\leq N
        """
        r_dash = self.unit_sphere_points[:, np.newaxis, :] - self.unit_sphere_points
        with np.errstate(divide="ignore", invalid="ignore"):
            single_forces = r_dash / (
                np.linalg.norm(r_dash, axis=2, keepdims=True) ** 3
            )
        return np.nansum(single_forces, axis=1)

    def _tangential_forces(self, forces):
        """
        Calculates the tangential part of the forces.

        First calculates the radial forces and substracts them to get the tangential part.
        With :math:`\\vec{r_i}` the radial vector pointing to point :math:`i`,
        :math:`F_i` the force on point :math:`i`, and scalar product :math:`\\langle~,~\\rangle`,
        the radial force can be calculated like:

            .. math::

                \\vec{F_{i_{rad}}} = \\langle\\vec{F_i}, \\vec{r_i}\\rangle  \\vec{r_i}

        The tangential force is then expressed by:

            .. math::

                \\vec{F_{i_{tan}}} = \\vec{F_i} - \\vec{F_{i_{rad}}}


        Parameters
        ----------
        forces : ndarray,
            The forces to calculate the tangential part of.

        """

        scalar_product = np.sum(forces * self.unit_sphere_points, axis=1, keepdims=True)
        radial_forces = scalar_product * self.unit_sphere_points
        return forces - radial_forces


if __name__ == "__main__":
    #    sphere = NSphere(3, 1000)
    #
    #    samples = sphere.unit_sphere_points
    #
    #    print("Initial : E_pot = {}".format(sphere.init_e_pot))
    #    print("Relaxed : E_pot = {}".format(sphere.rela_e_pot))
    #    print("Verbesserung: {} %".format(sphere.improvement))
    #
    #    #%% plot sphere
    #    from mpl_toolkits.mplot3d import Axes3D
    #    import matplotlib.pyplot as plt
    #    plt.close('all')
    #    fig = plt.figure(figsize=plt.figaspect(1))
    #    ax2 = fig.add_subplot(111, projection='3d')
    #
    #    u = np.linspace(0, 2 * np.pi, 50)
    #    v = np.linspace(0, np.pi, 30)
    #
    #    x = np.outer(np.cos(u), np.sin(v))
    #    y = np.outer(np.sin(u), np.sin(v))
    #    z = np.outer(np.ones(np.size(u)), np.cos(v))
    #
    #    ax2.plot_wireframe(x, y, z, rcount=10, ccount=10,)
    #
    #
    #    ax2.scatter(samples[:, 0], samples[:, 1], samples[:, 2],
    #                c='green', depthshade=False, label="verteilt")
    #
    #    ax2.legend()
    #    plt.show()
    import doctest

    doctest.testmod()



# This is based on the setup.py example at:
# https://github.com/pypa/sampleproject/blob/main/setup.py

from setuptools import setup

# To use a consistent encoding
import codecs
import os.path


long_description = """
virocon
=======

virocon is a software to compute environmental contours.

A longer description is available at: https://github.com/virocon-organization/virocon
"""


def read(rel_path):
    here = os.path.abspath(os.path.dirname(__file__))
    with codecs.open(os.path.join(here, rel_path), "r") as fp:
        return fp.read()


def get_version(rel_path):
    for line in read(rel_path).splitlines():
        if line.startswith("__version__"):
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1]
    else:
        raise RuntimeError("Unable to find version string.")


setup(
    name="virocon",
    version=get_version("virocon/__init__.py"),
    description="ViroCon is a software to compute environmental contours",
    long_description=long_description,
    url="https://github.com/virocon-organization/virocon",
    # We use git tags for this download_url. This approach is based on:
    # https://peterdowns.com/posts/first-time-with-pypi.html
    download_url="https://github.com/virocon-organization/virocon/archive/"
    + get_version("virocon/__init__.py")
    + ".tar.gz",
    author="virocon team",
    author_email="virocon@uni-bremen.de",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
    ],
    python_requires=">=3.7",
    packages=["virocon"],
    install_requires=[
        "matplotlib>=2.2.0",
        "networkx",
        "numpy",
        "pandas",
        "scikit-learn",
        "scipy",
    ],
    extras_require={
        "dev": ["Sphinx", "ruff", "pre-commit"],
        "test": ["pytest", "pytest-cov", "coverage", "coveralls"],
    },
    project_urls={  # Optional
        "Documentation": "https://virocon.readthedocs.io",
        "Source Code": "https://github.com/virocon-organization/virocon",
    },
)



import numpy as np

from virocon import (
    read_ec_benchmark_dataset,
    GlobalHierarchicalModel,
    WeibullDistribution,
    ExponentiatedWeibullDistribution,
    LogNormalDistribution,
    DependenceFunction,
    WidthOfIntervalSlicer,
    IFORMContour,
    HighestDensityContour,
    calculate_alpha,
    plot_marginal_quantiles,
    plot_dependence_functions,
    plot_2D_isodensity,
    plot_2D_contour,
)


def test_hs_tz_iform_contour():
    """
    Use a sea state dataset with the variables Hs and Tz,
    fit the join distribution recommended in DNVGL-RP-C203 to
    it and compute an IFORM contour. This tests reproduces
    the results published in Haseltseiner et al. (2019).

    Such a work flow is for example typical in ship design.

    Haselsteiner, A. F., Coe, R. G., Manuel, L., Nguyen, P. T. T.,
    Martin, N., & Eckert-Gallup, A. (2019). A benchmarking exercise
    on estimating extreme environmental conditions: Methodology &
    baseline results. Proc. 38th International Conference on Ocean,
    Offshore and Arctic Engineering (OMAE 2019).
    https://doi.org/10.1115/OMAE2019-96523

    DNV GL. (2017). Recommended practice DNVGL-RP-C205:
    Environmental conditions and environmental loads.
    """

    data = read_ec_benchmark_dataset("datasets/ec-benchmark_dataset_A.txt")

    # A 3-parameter power function (a dependence function).
    def _power3(x, a, b, c):
        return a + b * x**c

    # A 3-parameter exponential function (a dependence function).
    def _exp3(x, a, b, c):
        return a + b * np.exp(c * x)

    bounds = [(0, None), (0, None), (None, None)]
    power3 = DependenceFunction(_power3, bounds)
    exp3 = DependenceFunction(_exp3, bounds)

    dist_description_0 = {
        "distribution": WeibullDistribution(),
        "intervals": WidthOfIntervalSlicer(width=0.5),
    }
    dist_description_1 = {
        "distribution": LogNormalDistribution(),
        "conditional_on": 0,
        "parameters": {"mu": power3, "sigma": exp3},
    }
    model = GlobalHierarchicalModel([dist_description_0, dist_description_1])
    model.fit(data)

    plot_marginal_quantiles(model, data)
    plot_dependence_functions(model)
    plot_2D_isodensity(model, data)

    alpha = calculate_alpha(1, 20)
    contour = IFORMContour(model, alpha)

    coordinates = contour.coordinates
    np.testing.assert_allclose(max(coordinates[:, 0]), 5.0, atol=0.5)
    np.testing.assert_allclose(max(coordinates[:, 1]), 16.1, atol=0.5)

    plot_2D_contour(contour, sample=data)


def test_v_hs_hd_contour():
    """
    Use a wind speed - wave height dataset, fit the joint
    distribution that was proposed by Haselsteiner et al. (2020)
    and compute a highest density contour. This test reproduces
    the results presented in Haselestiner et al. (2020). The
    coorindates are availble at https://github.com/ec-benchmark-organizers/
    ec-benchmark/blob/main/results/exercise-1/contribution-4/haselsteiner_
    andreas_dataset_d_50.txt

    Such a work flow is for example typical when generationg
    a 50-year contour for DLC 1.6 in the offshore wind standard
    IEC 61400-3-1.

    Haselsteiner, A. F., Sander, A., Ohlendorf, J.-H., & Thoben, K.-D. (2020).
    Global hierarchical models for wind and wave contours: Physical
    interpretations of the dependence functions. Proc. 39th International
    Conference on Ocean, Offshore and Arctic Engineering (OMAE 2020).
    https://doi.org/10.1115/OMAE2020-18668

    International Electrotechnical Commission. (2019). Wind energy
    generation systems - Part 3-1: Design requirements for fixed
    offshore wind turbines (IEC 61400-3-1).
    """

    data = read_ec_benchmark_dataset("datasets/ec-benchmark_dataset_D.txt")

    def _logistics4(x, a=1, b=1, c=-1, d=1):
        return a + b / (1 + np.exp(c * (x - d)))

    def _alpha3(x, a, b, c, d_of_x):
        return (a + b * x**c) / 2.0445 ** (1 / d_of_x(x))

    logistics_bounds = [(0, None), (0, None), (None, 0), (0, None)]

    alpha_bounds = [(0, None), (0, None), (None, None)]

    beta_dep = DependenceFunction(_logistics4, logistics_bounds, weights=lambda x, y: y)
    alpha_dep = DependenceFunction(
        _alpha3, alpha_bounds, d_of_x=beta_dep, weights=lambda x, y: y
    )

    dist_description_v = {
        "distribution": ExponentiatedWeibullDistribution(),
        "intervals": WidthOfIntervalSlicer(width=2),
    }

    dist_description_hs = {
        "distribution": ExponentiatedWeibullDistribution(f_delta=5),
        "conditional_on": 0,
        "parameters": {
            "alpha": alpha_dep,
            "beta": beta_dep,
        },
    }

    model = GlobalHierarchicalModel([dist_description_v, dist_description_hs])

    fit_description_vs = {"method": "wlsq", "weights": "quadratic"}
    fit_description_hs = {"method": "wlsq", "weights": "quadratic"}

    model.fit(data, [fit_description_vs, fit_description_hs])

    plot_marginal_quantiles(model, data)
    plot_dependence_functions(model)
    plot_2D_isodensity(model, data)

    alpha = calculate_alpha(1, 50)
    limits = [(0, 35), (0, 20)]
    contour = HighestDensityContour(model, alpha, limits=limits, deltas=[0.2, 0.2])

    coordinates = contour.coordinates
    np.testing.assert_allclose(max(coordinates[:, 0]), 29.9, atol=0.2)
    np.testing.assert_allclose(max(coordinates[:, 1]), 15.5, atol=0.2)
    np.testing.assert_allclose(min(coordinates[:, 0]), 0, atol=0.1)
    np.testing.assert_allclose(min(coordinates[:, 1]), 0, atol=0.1)

    plot_2D_contour(contour, sample=data)



"""
A comprehensive example that shows the whole workflow of
1) Loading data
2) Defining the model structure for a joint distribution
3) Estimating the parameter values of the model ("Fitting")
4) Computing a 3D environmental contour

"""

import numpy as np
import matplotlib.pyplot as plt
import pandas as pd

from virocon import (
    GlobalHierarchicalModel,
    LogNormalDistribution,
    ExponentiatedWeibullDistribution,
    DependenceFunction,
    WidthOfIntervalSlicer,
    HighestDensityContour,
    plot_marginal_quantiles,
    plot_dependence_functions,
)


# Load sea state measurements.
data = pd.read_csv("datasets/coastDat2_oneyear.csv", sep=";", skipinitialspace=True)
data.index = pd.to_datetime(data.pop(data.columns[0]), format="%Y-%m-%d-%H")


# Define the structure of the joint model that we will use to describe
# the the environmental data. To define a joint model, we define the
# univariate parametric distributions and the dependence structure.
# The dependence structure is defined using parametric functions.


# A 3-parameter power function, which will be used as a dependence function.
def _power3(x, a, b, c):
    return a + b * x**c


# A 3-parameter exponential function, which will be used as a dependence function.
def _exp3(x, a, b, c):
    return a + b * np.exp(c * x)


# A 3- parameter alpha function, which will be used as a dependence function.
def _alpha3(x, a, b, c, d_of_x):
    return (a + b * x**c) / 2.0445 ** (1 / d_of_x(x))


# A 4- parameter logistic function, which will be used as a dependence function.
def _logistics4(x, a=1, b=1, c=-1, d=1):
    return a + b / (1 + np.exp(c * (x - d)))


# Lower and upper interval boundaries for the three parameter values.
bounds = [(0, None), (0, None), (None, None)]
logistics_bounds = [(0, None), (0, None), (None, 0), (0, None)]

power3 = DependenceFunction(_power3, bounds, latex="$a + b * x^c$")
exp3 = DependenceFunction(_exp3, bounds, latex="$a + b * \exp(c * x)$")
logistics4 = DependenceFunction(
    _logistics4,
    logistics_bounds,
    weights=lambda x, y: y,
    latex="$a + b / (1 + \exp[c * (x -d)])$",
)
alpha3 = DependenceFunction(
    _alpha3,
    bounds,
    d_of_x=logistics4,
    weights=lambda x, y: y,
    latex="$(a + b * x^c) / 2.0445^{1 / F()}$",
)

# Define the structure of the joint distribution.

# Wind speed.
dist_description_0 = {
    "distribution": ExponentiatedWeibullDistribution(),
    "intervals": WidthOfIntervalSlicer(2, min_n_points=50),
}
# Wave height.
dist_description_1 = {
    "distribution": ExponentiatedWeibullDistribution(f_delta=5),
    "intervals": WidthOfIntervalSlicer(0.5),
    "conditional_on": 0,
    "parameters": {
        "alpha": alpha3,
        "beta": logistics4,
    },
}
# Zero-up-crossing period.
dist_description_2 = {
    "distribution": LogNormalDistribution(),
    "conditional_on": 1,
    "parameters": {"mu": power3, "sigma": exp3},
}


model = GlobalHierarchicalModel(
    [dist_description_0, dist_description_1, dist_description_2]
)

# Define a dictionary that describes the model.
semantics = {
    "names": ["Wind speed", "Significant wave height", "Zero-up-crossing period"],
    "symbols": ["V", "H_s", "T_z"],
    "units": ["m/s", "m", "s"],
}

# Fit the model to the data (estimate the model's parameter values).
model.fit(data)

# Print the estimated parameter values.
print(model)

# Create plots to inspect the model's goodness-of-fit.
fig1, axs = plt.subplots(1, 3, figsize=[20, 7], dpi=300)
plot_marginal_quantiles(model, data, semantics, axes=axs)
fig2, axs = plt.subplots(1, 4, figsize=[20, 7], dpi=300)
plot_dependence_functions(model, semantics, axes=axs)

# Calculate 3D Contour.
state_duration = 1  # hours
return_period = 20  # years
alpha = state_duration / (return_period * 365.25 * 24)
HDC = HighestDensityContour(model, alpha, limits=[(0, 50), (0, 25), (0, 25)])

# randomly select only 5% of the contour's points to increase performance
rng = np.random.default_rng(42)
n_contour_points = len(HDC.coordinates)
random_points = rng.choice(
    n_contour_points, int(0.05 * n_contour_points), replace=False
)

Xs = HDC.coordinates[random_points, 0]
Ys = HDC.coordinates[random_points, 1]
Zs = HDC.coordinates[random_points, 2]

fig = plt.figure(dpi=300, figsize=[15, 7])
ax = fig.add_subplot(111, projection="3d")
ax.scatter(Xs, Ys, Zs, c="#004488")
ax.set_xlabel("Wind speed (m/s)")
ax.set_ylabel("Significant wave height (m)")
ax.set_zlabel("Zero-up-crossing period (s)")



"""
Dependence structures for joint distribution models.
"""

from inspect import signature
from functools import partial

from virocon._fitting import fit_function, fit_constrained_function

__all__ = ["DependenceFunction"]


# TODO test that order of execution does not matter
# it should not matter if the dependent or the conditioner are fitted first
class DependenceFunction:
    """
    Function to describe the dependencies between the variables.

    The dependence function is a function for the parameters of the dependent
    variable.

    Parameters
    ----------
    func : callable
        Dependence functions for the parameter.
        Maps a conditioning value x and an arbitrary number of parameters to
        the value of a distributions parameter y.
        func(x, \* args) -> y
    bounds : list
        Boundaries for parameters of func.
        Fixed scalar boundaries for func's parameters.
        E.g. 0 <= z <= 0 .
    constraints : dict
        More complex contraints modeled as unequality constraints with
        functions of the parameters of func z.
        I.e. c_j(z) >= 0 .
        For further explanation see:
        https://docs.scipy.org/doc/scipy-1.6.2/reference/tutorial/optimize.html#sequential-least-squares-programming-slsqp-algorithm-method-slsqp
    weights : callable, optional
        If given, weighted least squares fitting instead of least squares is
        used. Defaults to None.
        Given the data as observation tuples (x_i, y_i) maps from the vector
        x and y to the vector of weights.
        E.g. lambda x, y : y to linearly weight the observations with y_i.
    latex : string, optional
        If given, this string will be used in plots to label the dependence
        function. It is interpreted as latex and shoul be specified using the
        same symbols that are used in the function definition.
        Example: latex="$a + b \* x^{c}$"


    Examples
    --------
    The dependence function is a function for the parameters of the dependent
    variable. E.g.the zero-up-crossing period is dependent on the
    significant wave height (Hs|Tp). Assuming, the zero-upcrossing period is
    lognormally distributed, the parameters mu and sigma are described as
    functions of the significant wave height (equations given by
    Haselsteiner et. al(2020) [1]_ ).

    :math:`\\mu_{tz}(h_s) =  ln \\left(c_1 + c_2 \\sqrt{ \\frac{h_s)}{ g}} \\right)`

    :math:`\\sigma_{tz}(h_s) = c_3 + \\frac{c_4)}{1+ c_5h_s}`

    References
    ----------
    .. [1] Haselsteiner, A. F., Sander, A., Ohlendorf, J.-H., & Thoben, K.-D. (2020).
           Global hierarchical models for wind and wave contours: Physical interpretations
           of the dependence functions. Proc. 39th International Conference on Ocean,
           Offshore and Arctic Engineering (OMAE 2020). https://doi.org/10.1115/OMAE2020-18668
    """

    # TODO implement check of bounds and constraints
    def __init__(
        self, func, bounds=None, constraints=None, weights=None, latex=None, **kwargs
    ):
        # TODO add fitting method
        self.func = func
        self.bounds = bounds
        self.constraints = constraints
        self.weights = weights
        self.latex = latex

        # Read default values from function or set default as 1 if not specified.
        sig = signature(func)
        self.parameters = {
            par.name: (par.default if par.default is not par.empty else 1)
            for par in list(sig.parameters.values())[1:]
        }

        self.dependents = []

        self._may_fit = True
        self.dependent_parameters = {}
        self._fitted_conditioners = set()
        for key in kwargs.keys():
            if key in self.parameters.keys():
                self._may_fit = False
                dep_param = kwargs[key]
                self.dependent_parameters[key] = dep_param
                dep_param.register(self)
                dep_param_dict = {key: dep_param}
                self.func = partial(self.func, **dep_param_dict)
                del self.parameters[key]

    def __call__(self, x, *args, **kwargs):
        if len(args) + len(kwargs) == 0:
            return self.func(x, *self.parameters.values())
        elif len(args) + len(kwargs) == len(self.parameters):
            return self.func(x, *args, **kwargs)
        else:
            raise ValueError()  # TODO helpful error message

    def __repr__(self):
        if isinstance(self.func, partial):
            func = self.func.func
        else:
            func = self.func
        params = ", ".join(
            [
                f"{par_name}={par_value}"
                for par_name, par_value in self.parameters.items()
            ]
        )
        dep_params = ", ".join(
            [
                f"{par_name}={par_value}"
                for par_name, par_value in self.dependent_parameters.items()
            ]
        )
        combined_params = params + ", " + dep_params
        combined_params = combined_params.strip(", ")
        return f"DependenceFunction(func={func.__name__}, {combined_params})"

    def fit(self, x, y):
        """
        Determine the parameters of the dependence function.

        Parameters
        ----------

        x : array-like
            Input data (data consists of n observations (x_i, y_i)).
        y :array-like
            Target data (data consists of n observations (x_i, y_i)).

        Raises
        ------
        RuntimeError
            if the fit fails.

        """
        # The dependence function does not know in which order all the
        # dependence functions are fitted.
        # If another DependenceFunction has to be fitted before the current one,
        # the current one will not be fitted.
        # In the init the current dependence function registered at all
        # dependence functions which it depends on.
        # After fitting, every dependence functions signals all it's registered
        # dependence functions that it was fitted,
        # so that they know they may be fitted as well.

        # save x and y, this also marks that fit was called
        self.x = x
        self.y = y
        if self._may_fit:  # is the conditioner fitted, so that we can fit now?
            self._fit(self.x, self.y)

    def _fit(self, x, y):
        weights = self.weights
        if weights is not None:
            method = "wlsq"  # weighted least squares
            weights = weights(x, y)  # raises TypeError if not a callable
        else:
            method = "lsq"  # least squares

        bounds = self.bounds
        constraints = self.constraints

        # get initial parameters
        p0 = tuple(self.parameters.values())

        if constraints is None:
            try:
                popt = fit_function(self, x, y, p0, method, bounds, weights)
            except RuntimeError as e:
                raise RuntimeError(
                    f"Failed to fit dependence function {self}."
                    "Consider choosing different bounds."
                ) from e
        else:
            # TODO proper error handling for constrained fit
            popt = fit_constrained_function(
                self, x, y, p0, method, bounds, constraints, weights
            )

        # update self with fitted parameters
        self.parameters = dict(zip(self.parameters.keys(), popt))

        # after fitting inform dependents:
        for dependent in self.dependents:
            dependent.callback(self)

    def register(self, dependent):
        """
        Register a dependent DependenceFunction.
        The callback method of all registered dependents is called once this
        DependenceFunction was fitted.

        Parameters
        ----------
        dependent : DependenceFunction
            The DependenceFunctions to register.
        """

        self.dependents.append(dependent)

    def callback(self, caller):
        """
        Call to signal, that caller was fitted.

        Parameters
        ----------
        caller : DependeneFunction
            The DependenceFunctiom that is now fitted.
        """

        assert caller in self.dependent_parameters.values()
        # TODO raise proper error otherwise
        self._fitted_conditioners.add(caller)
        # check that all conditioners are already fitted, then we may fit self
        if self._fitted_conditioners.issubset(self.dependent_parameters.values()):
            self._may_fit = True
            if hasattr(self, "x") and hasattr(
                self, "y"
            ):  # did we try to fit earlier, but could not?
                self.fit(self.x, self.y)
